 
/* Intel IA32 instruction length decoder circuit generator
   Shaun Feng, UBC,  August, 2004
*/
  
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAX_PARCEL_SIZE      100 //maximum decoder buffer size
#define PAIR_SIZE        8 // one byte per read


#define instr_size 1000     // how many instructions?
                           // This number must be greater than
                           // the actual size of the instruction set

#define MAX_INSTR_LENGTH 17 


#define EMPTY -1      // Flags, didn't use them, but may use in the future
#define NONEMPTY 1    // if you don't have one byte size instruction,
                      // some set of

#define TEMP_VAR_PREFIX "_temp_"

                      // type for an instruction
struct instruction_{
  char* name;
  char* binary;
  int length;
  int op_length;
};
typedef struct instruction_* instr;


// list of the instructions with the same size.
struct size_list_{
  instr same_size_instr;
  struct size_list_* next;
};

// the set of (list instructions) with the same size
// it is the set of lists
struct set_list_{
  struct size_list_* set_table_node;
  struct set_list_* right;
  struct set_list_* next;
};



int PARCEL_SIZE = 32;
int instr_counter = 0;

struct instruction_ instr_array[instr_size];
struct size_list_* size_table[MAX_PARCEL_SIZE];
struct set_list_* set_table[MAX_PARCEL_SIZE];

int temp_var_table[MAX_PARCEL_SIZE][MAX_INSTR_LENGTH];

void init_all()
{
 int i,j;
 for(i = 0; i < PARCEL_SIZE; i++){
   size_table[i] = NULL;
   set_table[i] = NULL;
   for (j=0; j < MAX_INSTR_LENGTH; j++)
    temp_var_table[i][j] = 0;
 }
}

char* remove_colon(char* input){
  char *ret;
  int len = strlen(input)+1;
  int i,j=0;
  ret = (char*)malloc(sizeof(char)*len); 
  for (i=0; i<len-1; i++){
   if (input[i] !=':')
     ret[j++] = input[i]; 
  }
  ret[j] = '\0';
 return ret;
}

// read the input file, the format of that file is
// instruction_name      binary    instruction_length
// ex. ADDI             01        3

void read_input_file(char* input_file)
{

  FILE *input;
  char str1[20], str2[200];
  int len;
  struct size_list_ *temp, *newNode;

  init_all();
  input = fopen(input_file, "r");

  while(!feof(input)){
    fscanf(input, "%s %s %d\n", str1, str2, &len);
    instr_array[instr_counter].name = strdup(str1);
    
    //instr_array[instr_counter].binary = strdup(str2);
    instr_array[instr_counter].binary = remove_colon(str2);
    instr_array[instr_counter].length = len;
    instr_array[instr_counter].op_length =
         strlen(instr_array[instr_counter].binary);

    newNode = (struct size_list_*)malloc(sizeof(struct size_list_));
    newNode->same_size_instr = &instr_array[instr_counter];
    newNode->next = NULL;
    if (size_table[len-1] == NULL) size_table[len-1] = newNode;
    else {
           temp = size_table[len-1];
           while (temp->next != NULL) temp = temp->next;
           temp->next = newNode;
         }

    instr_counter++;
  }
}

void print_head()
{
  int i;
  printf("/**************************************************\n");
  printf("instruction length decoder\n");
  printf("%d instructions:\n",instr_counter);
  printf("instruction\t Opcode\t\t  Instruction_Length\n");
  for(i=0; i<instr_counter; i++){

    printf("%s\t\t %s\t\t  %d\n", instr_array[i].name ,
                            instr_array[i].binary,
                            instr_array[i].length);

  }
  printf("Copy Right. Shaun Feng, @UBC 2004\n");
  printf("***************************************************/\n");
  printf("\n`define PARCEL_SIZE\t %d\n", PARCEL_SIZE);
  printf("`define PAIR_SIZE\t %d\n", PAIR_SIZE);

}


void print_module_head()
{

  int i, first=0;
  printf("\nmodule main(c_wrapin, c_parcel, c_nextparcel, c_begin, c_end, c_wrapout);\n");
  printf("input [0: `PARCEL_SIZE - 1] c_wrapin;\n");
  printf("input [`PAIR_SIZE * `PARCEL_SIZE - 1: 0] c_parcel, c_nextparcel;\n");
  printf("\noutput[0: `PARCEL_SIZE - 1] c_begin, c_end, c_wrapout;\n");
/*
  printf("input c_wrapin, c_parcel, c_nextparcel;\n");
  printf("output c_begin, c_end, c_wrapout;\n\n");
  printf("wire[0: `PARCEL_SIZE - 1] c_wrapin, c_begin, c_end, c_wrapout;\n");
  printf("reg[`PAIR_SIZE * `PARCEL_SIZE - 1: 0] c_parcel, c_nextparcel;\n\n");
*/
}

// when we compute the size n, we can use the result of any i < n

//for size =1 to PARCEL_SIZE , do this procedure. assume we know size-1 set
void compute_set_size(int size)
{

  int temp_size = size, i;
  struct set_list_ *newNode, *temp;
  struct set_list_node_ *newItem;

  if (size_table[size-1] != NULL) {
       newNode =  (struct set_list_*)malloc(sizeof(struct set_list_));
       newNode->set_table_node = size_table[size-1];
       newNode->right = NULL;
       newNode->next  = NULL;
       if (set_table[size-1] == NULL) set_table[size-1] = newNode;
       else { 
              temp = set_table[size-1]->next;
              while (temp->next != NULL) temp = temp->next;
              temp->next = newNode;
            }
       
  } // if (instr_array[i].length == size) 

 //size >1 compute the set which size == size.
  for(i= size - 1; i> 0; i--){
         newNode =  (struct set_list_*)malloc(sizeof(struct set_list_));
         newNode->set_table_node =  size_table[i-1]; 
         newNode->right = set_table[size - i-1];
         newNode->next = NULL;
         if (size_table[i-1] == NULL || set_table[size-i-1] == NULL) {
           free(newNode);
         continue;
      }
      temp = set_table[size-1];
        if (temp == NULL) set_table[size-1] = newNode;
        else {
              while (temp->next != NULL) temp = temp->next;
              temp->next = newNode;
        }
  }// end for
 
}

void print_location_begin_end(int newlocation,
                              int nextParcel, 
                              int begin, 
                              int end,
                              char* str)
{
  if (nextParcel == 0){
  (begin==end)?
                          printf("c_parcel[%d] == %s",
                                  (newlocation+1) * PAIR_SIZE - begin - 1, str)
                          :
                          printf("c_parcel[%d : %d] == %d'b%s",
                                  (newlocation+1) * PAIR_SIZE - begin - 1,
                                  (newlocation+1) * PAIR_SIZE - end -1,
                                  end - begin +1, str);
  }
  if (nextParcel == 1){
   (begin==end)?
                          printf("c_nextparcel[%d] == %s",
                                  (newlocation+1) * PAIR_SIZE - begin - 1, str)
                          :
                          printf("c_nextparcel[%d : %d] == %d'b%s",
                                  (newlocation+1) * PAIR_SIZE - begin -1,
                                  (newlocation+1) * PAIR_SIZE - end -1,
                                  end - begin +1, str);
  }
}


// print_one_size_instr() must prior to read_one_size_instr();
void read_one_size_instr(int location, struct size_list_* length)
{
  if (length == NULL) return;
   // If already computed, return the result
   // temp_var_table is just a flag array. 1:Done 0:Unfinished
  if (temp_var_table[location][length->same_size_instr->length] == 1) {
     printf("%s%d_%d",
             TEMP_VAR_PREFIX,
             location,
             length->same_size_instr->length);
     return;
   }
}


void print_one_size_instr(int location, struct size_list_* length)
{
   // location is the position of PARCEL, not bit.   
  // size_table[length] has all the instructions with the same length == length

   int i, first, begin, end, state_flag, nextParcel, newLocation;
   struct size_list_* temp = length;
   char *p;
   char str[PAIR_SIZE*PARCEL_SIZE];
   int str_index=0;

   if (length == NULL) return;
   temp_var_table[location][length->same_size_instr->length] = 1;
   printf("wire %s%d_%d = ",
           TEMP_VAR_PREFIX,
           location,
           length->same_size_instr->length);
   printf("(");

   while(temp!= NULL){
             printf("(");
             p = temp->same_size_instr->binary;
             /*
             printf("whole p = %s, size = %d\n", p, temp->same_size_instr->length);
             for (i=0; i<temp->same_size_instr->op_length; i++){
               printf("\ntest p[%d] = %c\n",i, p[i]);
             }
             */
             first = 1; // print '&&' or not 
             state_flag = 0;
             // xxx101xxxx
             // state_flag = 0  *p is x or done one packet,
             // state_flag = 1  *p is 1/0
             newLocation = location;
             nextParcel = 0;

             for (i=0; i<temp->same_size_instr->op_length; i++) {

                  if (p[i] == '0' || p[i] == '1') {
                    // printf(" 0 or 1 \n ");
                     str[str_index++]=p[i];
                     if (state_flag == 0){
                         state_flag = 1;
                       // printf("state_flag   = 1\n");
                         // displacement inside one parcel
                         begin = end = i - (i/PAIR_SIZE)*PAIR_SIZE;
                     }
                     else
                         end++;
                  }

                  if (p[i] == 'x')  {
                     // xxx101xxxx
                     if (state_flag == 1){
                        str[str_index] = '\0';
                        if (first == 0) printf(" && ");
                        // print out from begin to end
                        print_location_begin_end(newLocation,
                                                 nextParcel,
                                                 begin,
                                                 end,
                                                 str);
                        first = 0;
                     }
                     // xxxx1111
                     state_flag  = 0;
                     str_index = 0;
                   }

                  // done one packet, need to print out.
                   //printf("\ni = %d, state_flag = %d\n",i,state_flag);
                  // xxxx1111 or
                  if ((i+1)%PAIR_SIZE == 0 && state_flag == 1){
                       str[str_index] = '\0';
                       if (first == 0) printf(" && ");
                       // print from begin to end
                       //printf("call print_location_begin_end \n");
                       print_location_begin_end(newLocation,
                                                nextParcel,
                                                begin,
                                                end,
                                                str);
                       first = 0;
                       state_flag =0;
                       str_index = 0;
                       begin = end = 0;
                  }
                  if ((i+1)%PAIR_SIZE == 0){
                       newLocation++;
                       if (newLocation > PARCEL_SIZE-1){
                         nextParcel = 1;
                         newLocation = 0;
                       }
                  }

             }//end of for
          printf(")");
          temp = temp->next;
          if (temp!=NULL) printf(" || ");
       }//end of while

       if (length!= NULL) printf(")");
       printf(";\n"); 
}



/* This is a recursive call to print all the traces of instructions with
total size of size, set = set_table[size-1], and start at location */
int print_or_size(int location, struct set_list_ *set)
{
// set_table[length] has all the set of items with a total length 
// struct set_list *temp = set, *temp2;

   if (set != NULL) {

     printf("(");
     read_one_size_instr(location, set->set_table_node);

     if (set->right!=NULL) { 
        printf("\n\t\t\t&&\n");
        print_or_size(location + set->set_table_node->same_size_instr->length,
                      set->right);
     }
     if (set->next!= NULL) {
        printf("\n\t\t  ||\n");
        print_or_size(location, set->next);
     }
  
     printf(")");
     return NONEMPTY;
  }
   return EMPTY;
}

/*
void print_begin()
{
  int i, j;
  for (i=0; i<PARCEL_SIZE; i++){
    printf("assign c_begin[%d] = c_wrapin[%d]", i, i);
    for (j=i-1; j >=0; j--){ 
        printf("\n\t|| ");
        printf("\n\t(c_begin[%d]&&", j);
        // instructions start at position j have total length i-j
 
        print_or_size(j,set_table[i - j - 1]); 
        printf(")");
     }
    printf(";\n");
  }
}
*/

void print_begin()
{
  int i, j;
  printf("\n");
  for (i=0; i<PARCEL_SIZE; i++){
    printf("assign c_begin[%d] = c_wrapin[%d]", i, i);
    for (j=i-1; j >=0; j--){
        if (size_table[i-j -1] != NULL) {
          printf("\n\t|| ");
          printf("\n\t(c_begin[%d]&&", j);
        // instructions start at position j have total length i-j
                                                                                                              
        //print_or_size(j,set_table[i - j - 1]);
          read_one_size_instr(j, size_table[i - j - 1]);
          printf(")");
        }
     }
    printf(";\n");
  }
}

void print_end()
{
  int i, j;
  printf("\n");                                                                                          
  for (i=0; i< PARCEL_SIZE; i++){
    if (i == PARCEL_SIZE-1)
      printf("assign c_end[%d] = ", i);
    else
      //printf("assign c_end[%d] = c_wrapin[%d]", i, i+1);
      {printf("assign c_end[%d] = c_begin[%d];\n",i, i+1); continue;}
                                                                                           
    for (j=i; j >=0; j--){
      if (size_table[i-j] != NULL) {
      if (j != PARCEL_SIZE -1) printf("\n\t|| \n\t");
       printf("(c_begin[%d]&&", j);
      //print_or_size(j,set_table[i - j]);
       read_one_size_instr(j, size_table[i - j]);
       printf(")");
      }
    }
    printf(";\n");
  }
}

void print_wrapout()
{
  int i, j, first;
  printf("\n");
  for (i=0; i< PARCEL_SIZE; i++){
    if (i == 0) {
      printf("assign c_wrapout[%d] = c_end[%d];\n",
              i, 
              PARCEL_SIZE-1);
      continue;
    }
    else
      printf("assign c_wrapout[%d] = ", i);
    first = 0;
    for (j=PARCEL_SIZE-1; j >=i; j--){
      if (size_table[PARCEL_SIZE - j + i -1] != NULL) {
         if (first == 1) printf("\n\t|| \n\t");
         printf("(c_begin[%d]&&", j);
         read_one_size_instr(j, 
                              size_table[PARCEL_SIZE - j + i -1]);
         first = 1;
         printf(")");
         //if (j>i) printf("||");
      }
    }
    if (first == 0) printf("0");

    printf(";\n");  
  }
} 

void set_compute()
{
  int i;
  for (i=1; i <= PARCEL_SIZE; i++)
   compute_set_size(i);
}

void print_module_tail()
{

  printf("endmodule\n");

}

void pre_assign(){
  int i,j;
  for (i =0; i< PARCEL_SIZE; i++)
   for (j=0; j< PARCEL_SIZE; j++)
     print_one_size_instr(i, size_table[j]);

}

int main(int argc, char* argv[])
{
  if (argc < 3) {
    printf("Format: command bufferSize instructionLengthTable\n");
    exit(-1);
  }
  PARCEL_SIZE = atoi(argv[1]);
  read_input_file(argv[2]);
  print_head();
  print_module_head();
  set_compute();
  pre_assign();
  print_begin();
  print_end();
  print_wrapout();
  print_module_tail();
}


