%{
/**************************************
*  The lex file for
*  IA32 instruction length decoder
**************************************
**************************************
*  Written by
*  Shaun Feng
*  April, 2005
**************************************/
    #include <stdio.h>
    #include <stdlib.h>
    #include <string.h>
    #include "parseia32.tab.h"
    extern int yyerror(char*);
    extern int yylineno;
    extern FILE *yyin;
%}
%start  instruction comment 

DIGIT    [0-9]
LETTER   [A-Za-z_]
ID       ({LETTER}({LETTER}|{DIGIT})*)
BINARYDIGIT [01xsw]
WS       [ \t\n\014\015:]
                                                                                     
%%
<instruction>"00001111"|"0000 1111"	{
			 yylval.string = strdup(yytext);
                         return ESCAPE;
			}
<instruction>"11110010"|"1111 0010"   {
                         yylval.string = strdup(yytext);
                         return MANDATORY_PREFIX;
                        }
<instruction>"11110011"|"1111 0011"   {
                         yylval.string = strdup(yytext);
                         return MANDATORY_PREFIX;
                        }
<instruction>"10101010"|"1010 1010"   {
                         yylval.string = strdup(yytext);
                         return MANDATORY_PREFIX;
                        }
<instruction>{BINARYDIGIT}{8}|{BINARYDIGIT}{4}" "{BINARYDIGIT}{4} {
			 yylval.string = strdup(yytext);
			 return BIT_8;		
                       }
<instruction>{BINARYDIGIT}{3} {
			 yylval.string = strdup(yytext);
			 return BIT_3;
                        }
<instruction>{BINARYDIGIT}{2} {
		 	 yylval.string = strdup(yytext);
			 return BIT_2;
                        }
<instruction>modA	{
                         yylval.string = strdup(yytext);
                         return MODA;
                        }
<instruction>mod        {
                         yylval.string = strdup(yytext);
                         return MOD;
                        }
<instruction>r\/m	{
                         return R_M;
			}
<instruction>reg	{
                         return REG;
			}
<instruction>immediate	{
                         return IMMEDIATE;
			}
<instruction>imm8       {
                         return IMM8;
                        }
<instruction>imm16      {
                         return IMM16;
                        }
<instruction>^["\\"|%]  {
                           BEGIN comment;
                        }
<instruction>{ID}	{
			  yylval.string = strdup(yytext);
			  return ID;
			}
<instruction>{WS}+      ;
<comment>(.)            ;       /* discard everything within a comment */
<comment>[\n]           {
                          BEGIN instruction; /* newline ends a comment */
                        }
                                                                                     
<instruction>.              { /* No other rule applies, so it's a bad token */
                             fprintf(stderr,"Bad input : %s \n",yytext);
                        }
%%

void
lex_init()
{
   BEGIN instruction;
}

int
yywrap()
{
 return 1;
}

int
lex_main( int argc, char **argv )
{
      ++argv, --argc;

      if ( argc > 0 )
               yyin = fopen( argv[0], "r" );
      else
               yyin = stdin;
      lex_init();                                                                                  
               yylex();
      return 0;
}
